#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

/*
  This gives a very cude C example that contours a "real world"
  dataset using the CONREC algorthm. This is by no means a general
  solution and uses many specific features of the particular data
  being contoured.
*/

/* The dimension of the grid */

int NX = 101;
int NY = 101;

#define M            10 
#define N            10 
#define Np           11 
#define Nm          490
#define Mwin        400
#define Nwin        400
#define Nl           10
#define Nd          ((N + 1) * (M + 1)) 
#define BOUND         0.0
#define ndata         3 * Mwin * Nwin

/* Table of constant values */

#define xdim       101
#define ydim       101
#define tdim       201
#define vdim       xdim * ydim
#define bdim       vdim * tdim * N

#define min(x, y)     ((x) < (y) ? (x) : (y))
#define max(x, y)     ((x) > (y) ? (x) : (y))
#define xsect(p1, p2) (h[p2] * xh[p1] - h[p1] * xh[p2]) / (h[p2] - h[p1])
#define ysect(p1, p2) (h[p2] * yh[p1] - h[p1] * yh[p2]) / (h[p2] - h[p1])

/* 
   The data will be scaled up by this amount to give some
   resolution for the contour lines.
*/

#define SCALE  4

// #define NMOVIE 10

/* Two dimensional array of data */

double **data;

/* b[2050401] was b[101][101][201] */

float b[bdim];	

/* Arrays for the x axis and y axis coordinates */

double *xaxis, *yaxis;

/* Array for the contour levels, 5 of them */

#define NCONTOUR 20

double contours[NCONTOUR];

/* Image on which the contours will be drawn, see bitmaplib.c */

/* ---> ERASED */

void draw(void);
void toolbox(double *x1, double *y1, double *x2, double *y2);

/* Prototype for CONREC and the line drawing function */

void CONREC(double **, int, int, int, int, double *, double *, int, double *,
	    void (*drawline)(double, double, double, double, double));
void drawline(double, double, double, double, double);

int potinc(float *b);
int convection(double rey, double *param, int nf, float *b);

void itoa(int n, char *s);
void reverse(char *s); 

/* Debugging - count the number of line segments drawn */

int vectorsdrawn = 0;

void expose(int width, int height)
{
  // Clear the window

  glClear(GL_COLOR_BUFFER_BIT);
}

void reshape(int width, int height)
{
  // Set the new viewport size

  glViewport(0, 0, (GLint)width, (GLint)height);

  // Choose the projection matrix to be the matrix 
  // manipulated by the following calls

  glMatrixMode(GL_PROJECTION);

  // Set the projection matrix to be the identity matrix

  glLoadIdentity();

  // Define the dimensions of the Orthographic Viewing Volume

  gluOrtho2D(- BOUND, (GLdouble) width, - BOUND, (GLdouble) height);
 
  // Choose the modelview matrix to be the matrix
  // manipulated by further calls

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Clear the window

  glClear(GL_COLOR_BUFFER_BIT);
}

void key_down(unsigned char key, int x, int y)
{
  switch(key) {
  case 27:
    exit(0);
    break;
  default:
    break;
  }
}

int runtime;
double Reynolds, mpar; 

/* Main program */

int main(int argc, char **argv)
{
  FILE *parametres;
  
  if((parametres = fopen("PAR", "r")) == NULL) {
    fprintf(stderr, "Failed to open data file \n");
    
    exit(- 1);
  }
  
  // parametres = fopen("PAR", "r");
  // fscanf(parametres, "%lf %lf %d", &Reynolds, &mpar, &runtime);
  
  if(fscanf(parametres, "%lf %lf %d", &Reynolds, &mpar, &runtime) != 3) {
    fprintf(stderr, "Enter the magnetic Reynolds number, "
	    "the m parametre, and the runtime \n");
    
    return(0);
  }
  
  fclose(parametres);
  
  // extern int potinc();
  
  glutInit(&argc, argv);
  glutInitWindowSize(Nwin, Mwin);
  glutInitWindowPosition(0, 0);

  // Open a window, name it "Points_Lines"

  glutCreateWindow(argv[0]);

  // Set the clear color to black

  // glClearColor(0.0, 0.0, 0.0, 0.0);
  
  // Set the clear color to white

  glClearColor(1.0, 1.0, 1.0, 1.0);

  // Set the shading model        

  glShadeModel(GL_FLAT);

  // Assign expose() to be the function called whenever   
  // an expose event occurs

  expose(N, M);

  // Assign reshape() to be the function called whenever 
  // a reshape event occurs

  reshape(N, M);

  glutReshapeFunc(reshape);
  
  // Assign key_down() to be the function called whenever 
  // a key is pressed

  glutKeyboardFunc(key_down);

  // potinc(a);

  // Assign draw() to be the function called whenever a display
  // event occurs, generally after a resize or expose event

  glutDisplayFunc(draw);
 
  // Assign idle() to be the function called whenever there               
  // are no events

  // glutIdleFunc(draw);
  
  // Pass program control to tk's event handling code
  // In other words, loop forever

  glutMainLoop();

  return(0); 
}

void draw(void)
{
  int i = 0, j, ii, jj, nn;
  int n = 0, lines = 0;
  int NMOVIE = runtime;
  double sum;
  double x, y, z;
  double xmin = 1e32, xmax = -1e32;
  double ymin = 1e32, ymax = -1e32;
  double zmin = 1e32, zmax = -1e32;
  double Reyn = Reynolds;
  double Parm = mpar;
  
  char *datos[ndata];
  
  char s1[20], s2[10], s3[5];

  FILE *pcell;
  short  TGA[] = {0, 2, 0, 0, 0, 0, Mwin, Nwin, 24};

  extern int potinc(float *b);
  extern int convection(double reyn, double *parm, int nf, float *b);

  /* 
     Malloc space for the two dimensional data 
     Initialise it to 0
  */
    
  data = malloc(((int)(SCALE * NX)) * sizeof(double *));

  for(i = 0; i < ((int) SCALE * NX); i++) {
    data[i] = malloc(((int)(SCALE * NY)) * sizeof(double));
  }

  for(i = 0; i < (int)(SCALE * NX); i++)
    for(j = 0; j < (int)(SCALE * NY); j++)
      data[i][j] = 0.0;

  /*

  for(i = 0; i < NX; i++)
    for(j = 0; j < NY; j++)
      for(n = 0; n < NX + NY; n++) {
	a[i + (j + n * NY) * NX] = 0.0;
      }

  **/

  /* 
     Open the data file and read the header.
     This is specific to the test dataset being used here,
     the header consists of the x and y dimensions of the data grid
     (which need not be fill)
  */

  nn = 0;

  potinc(b);

  for(nn = 1; nn <= NMOVIE; nn++) {
    /*
      Read the data.
      This is VERY specific to the details of this example data.
    */
    
    for(x = 0; x < (double) NX; x++) {
      for(y = 0; y < (double) NX; y++) {
	/* Spread the data over a block, for future smoothing */
	
	z = b[(int) x + ((int) y + nn * ydim) * xdim];
	
	for(ii = 0; ii < (int) SCALE; ii++)
	  for(jj = 0; jj < (int) SCALE; jj++)
	    data[((int)(SCALE * x)) + ii][((int)(SCALE * y)) + jj] = z;
	
	/* Check the bounds - debugging purposes */
	
	xmin = min(xmin, x);
	xmax = max(xmax, x);
	ymin = min(ymin, y);
	ymax = max(ymax, y);
	zmin = min(zmin, z);
	zmax = max(zmax, z);
	
	lines++;
      }
    }

    /*

    printf("Read %d lines \n", lines);
    printf("X Range: %lf -> %lf \n", xmin, xmax);
    printf("Y Range: %lf -> %lf \n", ymin, ymax);
    printf("Z Range: %lf -> %lf \n", zmin, zmax);
    
    **/

    /* 
       Smooth the data
       This is just a simple 4x4 rectangular filter, it isn't
       actually necessary but makes the result sexier. 
    */
    
    for(i = 0; i < (int)(SCALE * NX); i++) {
      for(j = 0; j < (int)(SCALE * NY); j++) {
	n = 0;
	sum = 0;
	
	for(ii = - 4; ii <= 4; ii++) {
	  for(jj = -4; jj <= 4; jj++) {
	    if(i + ii < 0 || i + ii >= (int)(SCALE * NX))
	      continue;
	    
	    if(j + jj < 0 || j + jj >= (int)(SCALE * NY))
	      continue;
	    
	    sum += data[i + ii][j + jj];
	    n++;
	  }
	}
	
	if(n <= 0) {
	  fprintf(stderr, "No cells averaged, this shouldn't happen! \n");
	  exit(- 1);
	}
	
	data[i][j] = sum / n;
      }
    }
    
    /* 
       Set up the axis coordinates 
       If helps to do this with thought so the line segments drawn
       by CONREC are in the most convenient coordinate system.
    */
    
    if((xaxis = malloc(((int)(SCALE * NX)) * sizeof(double))) == NULL) {
      fprintf(stderr, "Failed to malloc space for the xaxis data \n");
      exit(-1);
    }
    
    for(i = 0; i < ((int) SCALE * NX); i++) 
      xaxis[i] = i;
    
    if((yaxis = malloc(((int)(SCALE * NY)) * sizeof(double))) == NULL) {
      fprintf(stderr, "Failed to malloc space for the yaxis data \n");
      exit(- 1);
    }
    
    for(i = 0; i < (int)(SCALE * NY); i++) 
      yaxis[i] = i;
    
    /* Set up the contour levels */
    
    for(i = 0; i < NCONTOUR; i++) {
      contours[i] = zmin + (double) i / 20.0;  
    }
    
    /* 
       Create the image in memory and draw a colour ramped version 
       GetColour() just performs a colour ramp from blue to red.
    */
    
    /* ---> ERASED */
    
    /* Finally do the contouring */
    
    CONREC(data, 0, ((int)(SCALE * NX)) - 1, 0, ((int)(SCALE * NY)) - 1,
	   xaxis, yaxis, NCONTOUR, contours, drawline);

    // printf("Drew %d vectors \n", vectorsdrawn);

    glFinish();

    /* To save the Data */

    // glReadBuffer(GL_BACK);

    glReadBuffer(GL_FRONT);
    glReadPixels(0, 0, Mwin, Nwin, GL_RGB, GL_UNSIGNED_BYTE, datos);

    /* Writing concatenation */
    
    s1[0] = 'P';
    s1[1] = 'C';
    s1[2] = 'E';
    s1[3] = 'L';
    s1[4] = 'L';
    s1[5] = '\0';
    
    s3[0] = '.';
    s3[1] = 't';
    s3[2] = 'g';
    s3[3] = 'a';
    s3[4] = '\0';
    
    itoa(nn,  s2);
    strcat(s1, s2);
    strcat(s1, s3);
    
    printf("%s \n", s1);

    pcell = fopen(s1, "wb");

    fwrite(&TGA, sizeof(TGA), 1, pcell);
    fwrite(datos, ndata, 1, pcell);
    fclose(pcell);

    glutSwapBuffers();

    convection(Reyn, &Parm, nn, b);
    
    // Clear the window

    glClear(GL_COLOR_BUFFER_BIT);
    
    /* 
       Write the image as a TGA file 
       See bitmaplib.c for more details, or write "image"
       in your own prefered format.
    */
    
    /* ---> ERASED */
  }

  // exit(0);
}

void drawline(double x1, double y1, double x2, double y2, double z)
{
  /* ---> ERASED */

  // glLoadIdentity();

  // glColor3f(1.0, 1.0, 0.0);

  glColor3f(0.0, 0.0, 0.0);

  glTranslatef(0.0, 0.0, 0.0); 

  // Draw filtered lines

  // glEnable(GL_LINE_SMOOTH);

  // Set the line width

  glLineWidth(1.0);

  if(x1 < 0 || x1 >= NX * SCALE || x2 < 0 || x2 > NX * SCALE)
    fprintf(stderr, "Shouldn't get here, x out of bounds: %g %g \n", x1, x2);
  
  if(y1 < 0 || y1 >= NY * SCALE || y2 < 0 || y2 > NY * SCALE)
    fprintf(stderr, "Shouldn't get here, y out of bounds: %g %g \n", y1, y2);
  
  /* ---> ERASED */

  // Stippled line

  // glEnable(GL_LINE_STIPPLE);
  
  // glLineStipple(1, 0x0101); /* Dotted Line */

  glBegin(GL_LINES);
  glVertex2f(x1, y1);   
  glVertex2f(x2, y2);   
  glEnd();

  // Flush the buffer to force drawing of all objects thus far

  glFlush();

  // Swap the buffers

  glutSwapBuffers();

  vectorsdrawn++;
}

/*
  Derivation from CONREC
  d               ! matrix of data to contour
  ilb,iub,jlb,jub ! index bounds of data matrix
  x               ! data matrix column coordinates
  y               ! data matrix row coordinates
  nc              ! number of contour levels
  z               ! contour levels in increasing order
*/

void CONREC(double **d, int ilb, int iub, int jlb, int jub,
	    double *x, double *y, int nc, double *z,
	    void (*ConrecLine)(double, double, double, double, double))
{
  int m1, m2, m3, case_value;
  double dmin, dmax, x1 = 0.0, x2 = 0.0, y1 = 0.0, y2 = 0.0;
  int i, j, k, m;
  int sh[5];
  double h[5];
  double xh[5], yh[5];
  double temp1, temp2;
  int im[4] = {0, 1, 1, 0}, jm[4]={0, 0, 1, 1};
  int castab[3][3][3] = {{{0, 0, 8}, {0, 2, 5}, {7, 6, 9}},
			 {{0, 3, 4}, {1, 3, 1}, {4, 3, 0}},
			 {{9, 6, 7}, {5, 2, 0}, {8, 0, 0}}};
  
  for(j = (jub - 1); j >= jlb; j--) {
    for(i = ilb; i <= iub - 1; i++) {
      temp1 = min(d[i][j], d[i][j + 1]);
      temp2 = min(d[i + 1][j], d[i + 1][j + 1]);
      dmin  = min(temp1, temp2);
      temp1 = max(d[i][j], d[i][j + 1]);
      temp2 = max(d[i + 1][j], d[i + 1][j + 1]);
      dmax  = max(temp1, temp2);
      
      if(dmax < z[0] || dmin > z[nc-1])
	continue;
      
      for(k = 0; k < nc; k++) {
	if(z[k] < dmin || z[k] > dmax)
	  continue;
          
	for(m = 4; m >= 0; m--) {
	  if(m > 0) {
	    h[m]  = d[i + im[m - 1]][j + jm[m - 1]] - z[k];
	    xh[m] = x[i + im[m - 1]];
	    yh[m] = y[j + jm[m - 1]];
	  } else {
	    h[0]  = 0.25 * (h[1] + h[2] + h[3] + h[4]);
	    xh[0] = 0.50 * (x[i] +x[i + 1]);
	    yh[0] = 0.50 * (y[j] +y[j + 1]);
	  }
	  
	  if(h[m] > 0.0)
	    sh[m] = 1;
	  else if(h[m] < 0.0)
	    sh[m] = -1;
	  else
	    sh[m] = 0;
	}

	/*
	  Note: at this stage the relative heights of the corners and the
	  centre are in the h array, and the corresponding coordinates are
	  in the xh and yh arrays. The centre of the box is indexed by 0
	  and the 4 corners by 1 to 4 as shown below.
	  Each triangle is then indexed by the parameter m, and the 3
	  vertices of each triangle are indexed by parameters m1,m2,and m3.
	  It is assumed that the centre of the box is always vertex 2
	  though this isimportant only when all 3 vertices lie exactly on
	  the same contour level, in which case only the side of the box
	  is drawn.

	  vertex 4 +-------------------+ vertex 3
	           | \               / |
		   |   \    m-3    /   |
		   |     \       /     |
		   |       \   /       |
		   |  m=2    X   m=2   |       the centre is vertex 0
		   |       /   \       |
		   |     /       \     |
		   |   /    m=1    \   |
		   | /               \ |
	  vertex 1 +-------------------+ vertex 2
	
	*/
         
	/* Scan each triangle in the box */
	
	for(m = 1; m <= 4; m++) {
	  m1 = m;
	  m2 = 0;
	  
	  if(m != 4)
	    m3 = m + 1;
	  else
	    m3 = 1;
	  
	  if((case_value = castab[sh[m1] + 1][sh[m2] + 1][sh[m3] + 1]) == 0)
	    continue;
	  
	  switch(case_value) {
	  case 1: /* Line between vertices 1 and 2 */
	    x1 = xh[m1];
	    y1 = yh[m1];
	    x2 = xh[m2];
	    y2 = yh[m2];
	    break;
	  case 2: /* Line between vertices 2 and 3 */
	    x1 = xh[m2];
	    y1 = yh[m2];
	    x2 = xh[m3];
	    y2 = yh[m3];
	    break;
	  case 3: /* Line between vertices 3 and 1 */
	    x1 = xh[m3];
	    y1 = yh[m3];
	    x2 = xh[m1];
	    y2 = yh[m1];

	    break;
	  case 4: /* Line between vertex 1 and side 2-3 */
	    x1 = xh[m1];
	    y1 = yh[m1];
	    x2 = xsect(m2, m3);
	    y2 = ysect(m2, m3);

	    break;
	  case 5: /* Line between vertex 2 and side 3-1 */
	    x1 = xh[m2];
	    y1 = yh[m2];
	    x2 = xsect(m3, m1);
	    y2 = ysect(m3, m1);

	    break;
	  case 6: /* Line between vertex 3 and side 1-2 */
	    x1 = xh[m1];
	    y1 = yh[m1];
	    x2 = xsect(m1, m2);
	    y2 = ysect(m1, m2);
	    break;
	  case 7: /* Line between sides 1-2 and 2-3 */
	    x1 = xsect(m1, m2);
	    y1 = ysect(m1, m2);
	    x2 = xsect(m2, m3);
	    y2 = ysect(m2, m3);

	    break;
	  case 8: /* Line between sides 2-3 and 3-1 */
	    x1 = xsect(m2, m3);
	    y1 = ysect(m2, m3);
	    x2 = xsect(m3, m1);
	    y2 = ysect(m3, m1);

	    break;
	  case 9: /* Line between sides 3-1 and 1-2 */
	    x1 = xsect(m3, m1);
	    y1 = ysect(m3, m1);
	    x2 = xsect(m1, m2);
	    y2 = ysect(m1, m2);

	    break;
	  default:
	    break;
	  }
	  
	  /* Finally draw the line */
	
	  ConrecLine(x1, y1, x2, y2, z[k]);
	} /* m */
      } /* k - contour */
    } /* i */
  } /* j */
}

/* ====================================================================== */

/*     This is a sample vector output routine. For a local environment    */
/*     either replace the VECOUT call in the main line, or better         */
/*     replace the contents of this subroutine between the *'s shown.     */

/*     There is often the requirement to distinguish each contour         */
/*     line with a different colour or a different line style. This       */
/*     can be done in many ways using the contour values z for a          */
/*     particular line segment.                                           */

/* Subroutine */

int vecout(double *x1, double *y1, double *x2, double *y2, double *z)
{
  /* Subroutine */

  // extern void toolbx();
  
  /* ***** Replace from here */

  /*     The following should be ignored since it is specific to       */
  /*     the version of FORTRAN running on the Macintosh microcomputer */
  /*     on which this particular example was written.                 */

  /* toolbx(&x1, &y1, &x2, &y2); */

  glLoadIdentity();

  // glColor3f(1.0, 1.0, 0.0);

  glColor3f(0.0, 0.0, 0.0);

  glTranslatef(0.0, 0.0, 0.0); 

  // Draw filtered lines

  // glEnable(GL_LINE_SMOOTH);

  // Set the line width

  glLineWidth(1.0);

  // Stippled line

  // glEnable(GL_LINE_STIPPLE);
  
  // glLineStipple(1, 0x0101); /* Dotted Line */
  
  glBegin(GL_LINES);
  glVertex2f(*x1, *y1);   
  glVertex2f(*x2, *y2);   
  glEnd();

  // Flush the buffer to force drawing of all objects thus far

  glFlush();
  
  return(0);
} /* vecout */

void toolbox(double *x1, double *y1, double *x2, double *y2)
{
  glLoadIdentity();

  // glColor3f(1.0, 1.0, 0.0);  

  glColor3f(0.0, 0.0, 0.0);

  glTranslatef(0.0, 0.0, 0.0);

  // Draw filtered lines

  // glEnable(GL_LINE_SMOOTH);

  // Set the line width

  glLineWidth(1.0);

  // Stippled line

  // glEnable(GL_LINE_STIPPLE);
  
  // glLineStipple(1, 0x0101); /* Dotted Line */

  // glBegin(GL_LINE_STRIP);
  
  glBegin(GL_LINES);
  glVertex2f(*x1, *y1);   
  glVertex2f(*x2, *y2);   
  glEnd();  
}

/* Convection.c */

#define vx(i, j)   vx[(i) + (j) * xdim]
#define vy(i, j)   vy[(i) + (j) * xdim]
#define a(i, j, k) a[(i) + ((j) + (k) * ydim) * xdim]

/*
 * 
 * The Program Convect solves the induction equation for the vector 
 * potential by means of a stream function in a 2D region using the 
 * fourth order convective method
 *
 * Input parameters:
 *
 * Grid step size [DX, DY], number of steps [NX, NY]
 * Magnetic Reynolds number as a function of the resistivity, RE
 * RM = 1/RE
 * Time interval DT
 *
 */

/* Subroutine */ 

int convection(double rey, double *param, int nf, float *b)
{
  /* System generated locals */
  
  int i_1, i_2, i_3, i_4;
  double d_0;
  
  /* Builtin functions */

  double cos(double), sin(double);
  
  /* Local variables */

  static int i, j;
  static double l, m;
  static int n, r, s, u, z;
  static double ca, cb, cc, cd, fa, fe, ka, la, fi, na, md, mc, qa, ta;
  static int dr, cs, cu;
  static double re, dt, dx, dy, ex, ye, ms, mt, mu, qb, qc, qd, nx, ny;

  /* a[2050401] was [101][101][201] */
  
  static float a[bdim];	

  /* vx[10201] was vx[101][101] */ 
  /* vy[10201] was vy[101][101] */ 

  static float vx[vdim], vy[vdim], cda, cdb;
  static int cha, che, can;
  static double pad;
  static int cen;
  static double rad, qac, ped, qbc, red, qcc, pid, qdc, rid;
  static int con;
  static double rod, pau, rau, peu, reu, piu, riu, rou;
  static int crfx, crfy, crix, criy;

  dx = 0.01;
  dy = 0.01;
  dt = 0.01;
  nx = 1.0 / dx;
  ny = 1.0 / dy;
  re = 1.0 / rey;

  /* Input the initial values T_N and T_[N + 1/2] of the vector potential */
  
  i_1 = (int) nx; 
 
  for(i = 0; i <= i_1; i++) {
    i_2 = (int) ny;
    
    for(j = 0; j <= i_2; j++) {
      a(i, j, 0) = b[i + (j + (nf - 1) * ydim) * xdim];
      a(i, j, 1) = b[i + (j + nf * ydim) * xdim];
    }
  }
  
  /* Calculation of the plasma velocity components */

  i_1 = (int) nx;
  
  for(i = 0; i <= i_1; ++i) {
    i_2 = (int) ny;
    
    for(j = 0; j <= i_2; ++j) {
      /* Computing 2nd power */
	  
      d_0 = i * dx - 0.5;
      la = (1.0 - *param) * (d_0 * d_0) * 4.0 + *param;
      
      /* Computing 2nd power */
	    
      d_0 = j * dy - 0.5;
      ta = 1.0 - d_0 * d_0 * 4.0;
      ka = cos((i * dx - 0.5) * M_PI);
      na = sin((i * dx - 0.5) * M_PI);

      /* Computing 3rd power */
      
      d_0 = ta;

      vx(i, j) = la * (d_0 * (d_0 * d_0)) * (j * dy - 0.5) * ka 
	* (8.0 / M_PI);

      /* Computing 4th power */
      
      d_0 = ta, d_0 *= d_0;

      vy(i, j) = d_0 * d_0 
	* (- M_PI * la * na + (1.0 - *param) * 8.0 * ka * (i * dx - 0.5)) 
	* (1.0 / (4.0 * M_PI)); 
    }
  }

  /*
   * EX, YE, L and M are terms that appear frequently in the recurrence 
   * relations
   *
   */
  
  ex = dt / (dx * 12);
  ye = dt / (dy * 12);
  
  /* Computing 2nd power */
  
  d_0 = dx;
  l = re * dt / (d_0 * d_0);

  /* Computing 2nd power */
  
  d_0 = dy;
  m = re * dt / (d_0 * d_0);
  che = 2;
  
  /* N is the number of time intervals for the magnetic field evolution */
  
  for(n = 0; n <= 4; ++n) {
    r = 0;
    z = 1;

    /*
     * CHA measures if the current iteration is even (CHA = 0) or odd
     * (CHA = 1). It is use to control the sweep direction
     *
     */
    
    cha = n % che;
    
    /*
     * The grid sweep is defined as follows
     *
     * 1. X: from 0 to NX      Y: from 1 to NY - 1
     * 2. X: from 1 to NX - 1  Y: from 1 to NY - 1
     * 3. X: from 0 to NX      Y: from 2 to NY - 2
     * 4. X: from 0 to NX - 1  Y: from 2 to NY - 2
     *
     * In general
     *
     * X: from CRIX to CRFX    Y: from CRIY to CRFY
     *
     * Forward step example
     *
     * 0, 2, 4, ..., NX
     *
     * 1, 3, 5, ..., NY - 1 (first)
     *
     * The direction of the sweep is inverted in every time interval.
     * DR controls this invertion.
     *
     */

    for(con = 1; con <= 2; ++con) {
      u = 0;
      
      for(cen = 1; cen <= 2; ++cen) {
	if(cha == 0) {
	  cs = 1;
	  criy = r + 1;
	  crfy = (int) (ny - 1 - r);
	  dr = 2;
	  crix = u;
	  crfx = (int) (nx - u);
	} else {
	  cs = - 1;
	  criy = (int) (ny - 1 - r);
	  crfy = r + 1;
	  crix = (int) (nx - u);
	  crfx = u;
	  dr = - 2;
	}

	/* Calculation of the vector potential */

	i_1 = crfy;
	i_2 = dr;
	
	for(j = criy; i_2 < 0 ? j >= i_1 : j <= i_1; j += i_2) {
	  i_3 = crfx;
	  i_4 = dr;
	  
	  for(i = crix; i_4 < 0 ? i >= i_3 : i <= i_3; i += i_4) {
	    fa = 0.0;
	    
	    /*
	     * The potential and velocity coordinates in each grid point are 
	     * adjusted by means of parameters. Defining the variables as 
	     * function of certain symmetries each term of the potential can 
	     * be decomposed in two simmetric parts which can be calculated 
	     * from one equation. The parameters S and Z are adjusted 
	     * according to the calculated term and the sweep direction.
	     *
	     */

	    for(can = 1; can <= 2; ++can) {
	      if(can == 1) {
		s = -1;
	      } else {
		s = 1;
	      }
		
	      if(cha == 0) {
		if(can == 1) {
		  z = 1;
		} else {
		  z = 0;
		}
	      } else {
		if(can == 2) {
		  z = 1;
		} else {
		  z = 0;
		}
	      }
	      
	      cu = 0;
	      mu = (double) (i + s);
	      md = (double) (i - 2);
	      mt = (double) (i + 2);
	      mc = (double) (j - 2);
	      ms = (double) (j + 2);
	      qac = mt - 1;
	      qbc = md + 1;
	      qcc = ms - 1;
	      qdc = mc + 1;
	      
	      /* 
	       * The values of each coordinate are assigned according to 
	       * the boundary conditions. 
	       *
	       */

	      if((double) i == nx && s == 1) {
		mu = 0.0;
		cu = 1;
	      }
	      
	      if(i == 0 && s == -1) {
		mu = 0.0;
		cu = 2;
	      }
		
	      if(i == 0) {
		md = 0.0;
		qbc = 0.0;
	      }
		
	      if(i == 1) {
		md = 0.0;
	      }
			
	      if((double) i == nx) {
		mt = 0.0;
		qac = 0.0;
	      }
			 
	      if((double) i == nx - 1) {
		mt = 0.0;
	      }
		
	      if(j == 1) {
		mc = 0.0;
		qdc = 0.0;
		
		if((double) i != nx && s != 1) {
		  if (i != 0 && s != - 1) {
		    cu = 3;
		  }
		}
	      }
		
	      if((double) j == ny - 1) {
		ms = 0.0;
		qcc = ny;
		
		if((double) i != nx && i != 1) {
		  if(i != 0 && i != - 1) {
		    cu = 4;
		  }
		}
	      }

	      /* Values of each term used to calculate the potential */
	      
	      pau = vx((int) mu, j);
	      peu = vx((int) mu, (int) ms);
	      piu = vx((int) mu, (int) mc);
	      pad = vy(i, (j + s));
	      ped = vy((int) mt, (j + s));
	      pid = vy((int) md, (j + s));
	      rau = a((int) mu, j, (n + 1));
	      reu = a((int) md, j, (n + z));
	      riu = a((int) mu, (int) ms, (n + 1));
	      rou = a((int) mu, (int) mc, (n + 1));
	      rad = a(i, (j + s), (n + 1));
	      red = a(i, (int) mc, (n + z));
	      rid = a((int) mt, (j + s), (n + 1));
	      rod = a((int) md, (j + s), (n + 1));
	      qa = a((int) qac, j, (n + 1));
	      qb = a((int) qbc, j, (n + 1));
	      qc = a(i, (int) qcc, (n + 1));
	      qd = a(i, (int) qdc, (n + 1));
	      
	      /* Values of these term at the boundaries */

	      if(cu == 1) {
		pau = vx((int) (nx - 1), j) - vy((int) nx, (j + 1)) 
		  + vy((int) nx, (j - 1));
		
		if((double) j != ny - 1 && (double) j != ny - 2) {
		  peu = vx((int) (nx - 1), (j + 2)) - vy((int) nx, (j + 3)) 
		    + vy((int) nx, (j + 1));
		}
		
		if(j != 1 && j != 2) {
		  piu = vx((int) (nx - 1), (j - 2)) - vy((int) nx, (j - 1)) 
		    + vy((int) nx, (j - 3));
		}
		
		if((double) j == ny - 2) {
		  peu = vx((int) (nx - 1), (int) ny);
		}
		
		if(j == 2) {
		  piu = vx((int) (nx - 1), 0);
		}
		
		if(j == 1 || (double) j == ny - 1) {
		  peu = 0.0;
		  piu = 0.0;
		}
		
		rau = 8.99f;
		riu = 8.99f;
		rou = 8.99f;
	      }
		
	      if(cu == 2) {
		pau = vx(1, j) + vy(0, (j + 1)) - vy(0, (j - 1));
		
		if((double) j != ny - 1 && (double) j != ny - 2) {
		  peu = vx(1, (j + 2)) + vy(0, (j + 3)) - vy(0, (j + 1));
		}
		
		if(j != 1 && j != 2) {
		  piu = vx(1, (j - 2)) + vy(0, (j - 1)) - vy(0, (j - 3));
		}
		
		if((double) j == ny - 2) {
		  peu = vx(1, (int) ny);
		}
		
		if(j == 2) {
		  piu = vx(1, 0);
		}
		
		if(j == 1 || (double) j == ny - 1) {
		  peu = 0.0;
		  piu = 0.0;
		}
				
		rau = 10.01f;
		riu = 10.01f;
		rou = 10.01f;
	      }
	      
	      if(i == 0) {
		pid = ped;
		reu = 10.02f;
		rod = 10.02f;
		qb = 10.01f;
	      }
		
	      if(i == 1) {
		pid = ped;
		reu = 10.01f;
		rod = 10.01f;
	      }
	      
	      if((double) i == nx) {
		ped = pid;
		rid = 8.98f;
		qa = 8.99f;
	      }
	      
	      if((double) i == nx - 1) {
		ped = pid;
		rid = 8.99f;
	      }
		
	      if(cu == 3) {
		piu = 0.0;
		rou = a((int) mu, 1, (n + 1));
		red = a(i, 1, (n + z));
	      }
	      
	      if(cu == 4) {
		peu = 0.0;
		riu = a((int) mu, (int) (ny - 1), (n + 1));
	      }
	      
	      /* Calculation of the potencial */

	      ca = - s * ex * (pau * (rau * 8 - reu) 
			       + (riu - rou) * (peu - piu) / 8);
	      cb = - s * ye * (pad * (rad * 8 - red) 
			       + (rid - rod) * (ped - pid) / 8);
	      fa = fa + ca + cb;
	      cd = 1 - cs * (ex * pau + ye * pad) + cs * s * (l + m);

	      if(cs == 1) {
		if(s == - 1) {
		  cda = cd;
		} else {
		  cdb = cd;
		}
	      }
	      
	      if(cs == - 1) {
		if(s == 1) {
		  cda = cd;
		} else {
		  cdb = cd;
		}
	      }
	    }
		
	    cc = l * (qa + qb) + m * (qc + qd);
	    fe = fa + cc;
	    fi = fe + a(i, j, n) * cda;
	    a(i, j, (n + 2)) = fi / cdb;
	  }
	}
	
	u = 1;
      }
	
      r = 1;
    }

    /*
     * Calculation of the potential values that remain fixed 
     * (lower and upper row)
     *
     */
      
    i_2 = (int) nx;
	
    for(i = 0; i <= i_2; ++i) {
      a(i, 0, (n + 2)) = a(i, 0, n);
      a(i, (int) ny, (n + 2)) = a(i, (int) ny, n);
    }
  }
    
  /* Storage of the new potential values at times T_N and T_[N - 1] */

  i_1 = (int) nx;

  for(i = 0; i <= i_1; i++) {
    i_2 = (int) ny;

    for(j = 0; j <= i_2; j++) {
      b[i + (j + nf * ydim) * xdim]       = a(i, j, n + 1);      
      b[i + (j + (nf + 1) * ydim) * xdim] = a(i, j, n + 1);
    }
  }

  return(0);
} /* convection */

/* Potinc.c */

/* Table of constant values */

/* adim = idim * jdim * kdim = 111 * 111 * 51 = 628371 */

#define idim        111
#define jdim        111
#define kdim        51
#define adim        idim * jdim * kdim
#define an(i, j, k) an[(i) + ((j) + (k) * kdim) * jdim]

int potinc(float *b)
{
  /* System generated locals */

  float r_1, r_2;

  /* Local variables */
  
  /* a[628371] was [111][111][51] */

  static float an[adim], c, i, j, dx, dy, nx, ny;

  dx = 0.01;
  dy = 0.01;
  c  = 1.0;

  an(0, 0, 0) = 10.0;
  an(0, 1, 0) = 10.0;
  an(1, 0, 0) = an(0, 0, 0) - dx * c;
  an(1, 1, 0) = an(0, 1, 0) - dx * c;

  nx = 1.0 / dx;
  ny = 1.0 / dy;

  for(j = 0.0; j <= 1.0; j += 1.0) {
    r_1 = nx;
    
    for(i = 0.0; i <= r_1; i += 2.0) {
      an((int) i, (int) j, 0) = an(0, (int) j, 0) - i * dx * c;
    }
    
    r_1 = nx;
	
    for(i = 1.0; i <= r_1; i += 2.0) {
      an((int) i, (int) j, 0) = an(1, (int) j, 0) - (i - 1) * dx * c;
    }
  }
  
  r_1 = nx;
  
  for(i = 0.0; i <= r_1; i += 1.0) {
    r_2 = ny;
    
    for(j = 2.0; j <= r_2; j += 2.0) {
      an((int) i, (int) j, 0) = an((int) i, 0, 0);
    }
    
    r_2 = ny;
	
    for(j = 1.0; j <= r_2; j += 2.0) {
      an((int) i, (int) j, 0) = an((int) i, 1, 0 );
    }
  }

  r_1 = nx;
  
  for(i = 0.0; i <= r_1; i += 1.0) {
    r_2 = ny;
    
    for(j = 0.0; j <= r_2; j += 1.0) {
      b[(int) i + ((int) j) * ydim]        = an((int) i, (int) j, 0);
      b[(int) i + ((int) j + ydim) * xdim] = an((int) i, (int) j, 0);
    }
  }

  return(0);
} 

/* itoa: n in Text in s wandeln -- Variante */

void itoa(int n, char *s)
{
  int i, sign;
  
  /* Vorzeichen notieren */

  sign = n;
  i = 0;

  /* Ziffern in ungekehrter Reihenfolge */

  do {
    /* naechster Ziffer wegdividieren */

    s[i++] = abs(n % 10) + '0';
  } while((n /= 10) != 0);
  
  if(sign < 0) 
    s[i++] = '-';

  s[i] = '\0';

  reverse(s);
}

/* reverse: Zeichenkette s umkehren */

void reverse(char *s) 
{
  int i, j;
  char temp;

  i = 0;

  /* Ende von s finden */

  while(s[i] != '\0')
    ++i;

  /* ein Element vor "\0" */

  --i;

  /* Zeilentrenner stehenlassen */

  if(s[i] == '\n')
    --i;

  /* Anfang von s */

  j = 0; 

  while(j < i) {
    temp = s[j];

    /* Zeichen tauschen */

    s[j] = s[i];
    s[i] = temp;
    
    --i;
    ++j;
  }
}
