/* Stream.f -- translated by f2c (version 19991025). */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

/*
  This gives a very cude C example that contours a "real world"
  dataset using the CONREC algorthm. This is by no means a general
  solution and uses many specific features of the particular data
  being contoured.
*/

/* The dimension of the grid */

int NX = 101;
int NY = 101;

#define M            10 
#define N            10 
#define Np           11
#define xdim        101 
#define Nm          490
#define Mwin        400
#define Nwin        400
#define Nl           10
#define Nd          ((N + 1) * (M + 1)) 
#define BOUND         0.0
#define gdim        101
#define stdim       gdim * gdim
#define st(i, j)    st[((i) - 1) + ((j) - 1) * gdim]

#define min(x, y)     ((x) < (y) ? (x) : (y))
#define max(x, y)     ((x) > (y) ? (x) : (y))
#define xsect(p1, p2) (h[p2] * xh[p1] - h[p1] * xh[p2]) / (h[p2] - h[p1])
#define ysect(p1, p2) (h[p2] * yh[p1] - h[p1] * yh[p2]) / (h[p2] - h[p1])

/* 
   The data will be scaled up by this amount to give some
   resolution for the contour lines.
*/

#define SCALE 4

/* Two dimensional array of data */

double **data;

/* Arrays for the x axis and y axis coordinates */

double *xaxis, *yaxis;

/* Array for the contour levels, 5 of them */

#define NCONTOUR 40

double contours[NCONTOUR];

/* Image on which the contours will be drawn, see bitmaplib.c */

/* ---> ERASED */

void draw(void);
int stream(double pm);

/* Prototype for CONREC and the line drawing function */

void CONREC(double **, int, int, int, int, double *, double *, int, double *,
	    void (*drawline)(double, double, double, double, double));
void drawline(double, double, double, double, double);

/* Debugging - count the number of line segments drawn */

int vectorsdrawn = 0;

/* st[10201] was st[101][101] */

static float st[stdim];	

void expose(int width, int height)
{
  // Clear the window

  glClear(GL_COLOR_BUFFER_BIT);
}

void reshape(int width, int height)
{
  // Set the new viewport size

  glViewport(0, 0, (GLint)width, (GLint)height);

  // Choose the projection matrix to be the matrix 
  // manipulated by the following calls

  glMatrixMode(GL_PROJECTION);

  // Set the projection matrix to be the identity matrix

  glLoadIdentity();

  // Define the dimensions of the Orthographic Viewing Volume

  gluOrtho2D(- BOUND, (GLdouble) width, - BOUND, (GLdouble) height);
 
  // Choose the modelview matrix to be the matrix
  // manipulated by further calls

  glMatrixMode(GL_MODELVIEW);

  // Clear the window

  glClear(GL_COLOR_BUFFER_BIT);
}

void key_down(unsigned char key, int x, int y)
{
  switch(key) {
  case 27:
    exit(0);
    break;
  }
}

/* Main program */

int main(int argc, char **argv)
{
  glutInit(&argc, argv);
  glutInitWindowSize(Nwin, Mwin);
  glutInitWindowPosition(0, 0);

  // Open a window, name it "Points_Lines"

  glutCreateWindow(argv[0]);

  // Set the clear color to black

  // glClearColor(0.0, 0.0, 0.0, 0.0);

  // Set the clear color to white

  glClearColor(1.0, 1.0, 1.0, 1.0);
  
  // Set the shading model        

  glShadeModel(GL_FLAT);
  
  // Assign expose() to be the function called whenever   
  // an expose event occurs

  expose(N, M);

  // Assign reshape() to be the function called whenever 
  // a reshape event occurs

  glutReshapeFunc(reshape);
  
  // Assign key_down() to be the function called whenever 
  // a key is pressed

  glutKeyboardFunc(key_down);
        
  // Assign draw() to be the function called whenever a display
  // event occurs, generally after a resize or expose event

  glutDisplayFunc(draw);

  // Pass program control to tk's event handling code
  // In other words, loop forever

  glutMainLoop();

  return 0; 
}

void draw(void)
{
  int i, j, ii, jj;
  int n = 0, lines = 0;
  double sum;
  double x, y, z;
  double xmin = 1e32, xmax = -1e32;
  double ymin = 1e32, ymax = -1e32;
  double zmin = 1e32, zmax = -1e32;
  double param;

  extern int stream();

  /* 
     Open the data file and read the header.
     This is specific to the test dataset being used here,
     the header consists of the x and y dimensions of the data grid
     (which need not be fill)
  */

  FILE *streamdat;

  /* 
     Malloc space for the two dimensional data 
     Initialise it to 0
  */
 
  if((data = malloc(((int)(SCALE * NX)) * sizeof(double *))) == NULL) {
    fprintf(stderr, "Failed to malloc space for the data \n");
    exit(-1);
  }
   
  for(i = 0; i < ((int) SCALE * NX); i++) {
    if((data[i] = malloc(((int)(SCALE * NY)) * sizeof(double))) == NULL) {
      fprintf(stderr, "Failed to malloc space for the data \n");
      exit(-1);
    }
  }
  
  for(i = 0; i < (int)(SCALE * NX); i++)
    for(j = 0; j < (int)(SCALE * NY); j++)
      data[i][j] = 0.0;
  
  /*
    Read the data.
    This is VERY specific to the details of this example data.
  */

  streamdat = fopen("ST", "r");
  
  fscanf(streamdat, "%lf", &param); 
  fclose(streamdat);
  
  stream(param);

  /* Spread the data over a block, for future smoothing */

  for(x = 0; x < (double) NX; x++) {
    for(y = 0; y < (double) NX; y++) {
      /* Spread the data over a block, for future smoothing */
      
      z = st[(int) x + ((int) y ) * xdim];
      
      for(ii = 0; ii < (int) SCALE; ii++)
	for(jj = 0; jj < (int) SCALE; jj++)
	  data[((int)(SCALE * x)) + ii][((int)(SCALE * y)) + jj] = z;
      
      /* Check the bounds - debugging purposes */
      
      xmin = min(xmin, x);
      xmax = max(xmax, x);
      ymin = min(ymin, y);
      ymax = max(ymax, y);
      zmin = min(zmin, z);
      zmax = max(zmax, z);
      
      lines++;
    }
  }  
  
  /*
    fprintf(stderr, "Read %d lines \n", lines);
    fprintf(stderr, "X Range: %lf -> %lf \n", xmin, xmax);
    fprintf(stderr, "Y Range: %lf -> %lf \n", ymin, ymax);
    fprintf(stderr, "Z Range: %lf -> %lf \n", zmin, zmax);
  **/
  
  /* 
     Smooth the data
     This is just a simple 4x4 rectangular filter, it isn't
     actually necessary but makes the result sexier. 
  */
  
  for(i = 0; i < (int)(SCALE * NX); i++) {
    for(j = 0; j < (int)(SCALE * NY); j++) {
      n = 0;
      sum = 0;
      
      for(ii = - 4; ii <= 4; ii++) {
	for(jj = -4; jj <= 4; jj++) {
	  if(i + ii < 0 || i + ii >= (int)(SCALE * NX))
	    continue;
	  
	  if(j + jj < 0 || j + jj >= (int)(SCALE * NY))
	    continue;
	  
	  sum += data[i + ii][j + jj];
	  n++;
	}
      }
      
      if(n <= 0) {
	fprintf(stderr, "No cells averaged, this shouldn't happen! \n");
	exit(-1);
      }
      
      data[i][j] = sum / n;
    }
  }
  
  /* 
     Set up the axis coordinates 
     If helps to do this with thought so the line segments drawn
     by CONREC are in the most convenient coordinate system.
  */
  
  if((xaxis = malloc(((int)(SCALE * NX)) * sizeof(double))) == NULL) {
    fprintf(stderr, "Failed to malloc space for the xaxis data \n");
    exit(-1);
  }
  
  for(i = 0; i < ((int) SCALE * NX); i++) 
    xaxis[i] = i;
  
  if((yaxis = malloc(((int)(SCALE * NY)) * sizeof(double))) == NULL) {
    fprintf(stderr, "Failed to malloc space for the yaxis data \n");
    exit(-1);
  }
  
  for(i = 0; i < (int)(SCALE * NY); i++) 
    yaxis[i] = i;
  
  /* Set up the contour levels */
  
  for(i = 0; i < NCONTOUR; i++) {
    contours[i] = zmin + (double) i / 40.0;  
  }
  
  /* 
     Create the image in memory and draw a colour ramped version 
     GetColour() just performs a colour ramp from blue to red.
  */
  
  /* ---> ERASED */
  
  /* Finally do the contouring */
  
  CONREC(data, 0, ((int)(SCALE * NX)) - 1, 0, ((int)(SCALE * NY)) - 1,
	 xaxis, yaxis, NCONTOUR, contours, drawline);
  
  // fprintf(stderr, "Drew %d vectors \n", vectorsdrawn);
  
  /* 
     Write the image as a TGA file 
     See bitmaplib.c for more details, or write "image"
     in your own prefered format.
  */
  
  /* ---> ERASED */
}

void drawline(double x1, double y1, double x2, double y2, double z)
{
  /* ---> ERASED */

  glLoadIdentity();

  // glColor3f(1.0, 1.0, 0.0);

  glColor3f(0.0, 0.0, 0.0);

  glTranslatef(0.0, 0.0, 0.0); 

  // Draw filtered lines

  // glEnable(GL_LINE_SMOOTH);

  // Set the line width

  glLineWidth(1.0);

  if(x1 < 0 || x1 >= NX * SCALE || x2 < 0 || x2 > NX * SCALE)
    fprintf(stderr, "Shouldn't get here, x out of bounds: %g %g \n", x1, x2);
  
  if(y1 < 0 || y1 >= NY * SCALE || y2 < 0 || y2 > NY * SCALE)
    fprintf(stderr, "Shouldn't get here, y out of bounds: %g %g \n", y1, y2);
  
  /* ---> ERASED */

  // Stippled line

  // glEnable(GL_LINE_STIPPLE);
  
  // glLineStipple(1, 0x0101); /* Dotted Line */

  glBegin(GL_LINES);
  glVertex2f(x1, y1);   
  glVertex2f(x2, y2);   
  glEnd();

  // Flush the buffer to force drawing of all objects thus far

  glFlush();
  
  vectorsdrawn++;
}

/*
  Derivation from CONREC
  d               ! matrix of data to contour
  ilb,iub,jlb,jub ! index bounds of data matrix
  x               ! data matrix column coordinates
  y               ! data matrix row coordinates
  nc              ! number of contour levels
  z               ! contour levels in increasing order
*/

void CONREC(double **d, int ilb, int iub, int jlb, int jub,
	    double *x, double *y, int nc, double *z,
	    void (*ConrecLine)(double, double, double, double, double))
{
  int m1, m2, m3, case_value;
  double dmin, dmax, x1 = 0.0, x2 = 0.0, y1 = 0.0, y2 = 0.0;
  int i, j, k, m;
  int sh[5];
  double h[5];
  double xh[5], yh[5];
  double temp1, temp2;
  int im[4] = {0, 1, 1, 0}, jm[4]={0, 0, 1, 1};
  int castab[3][3][3] = {{{0, 0, 8}, {0, 2, 5}, {7, 6, 9}},
			 {{0, 3, 4}, {1, 3, 1}, {4, 3, 0}},
			 {{9, 6, 7}, {5, 2, 0}, {8, 0, 0}}};
  
  for(j = (jub - 1); j >= jlb; j--) {
    for(i = ilb; i <= iub - 1; i++) {
      temp1 = min(d[i][j], d[i][j + 1]);
      temp2 = min(d[i + 1][j], d[i + 1][j + 1]);
      dmin  = min(temp1, temp2);
      temp1 = max(d[i][j], d[i][j + 1]);
      temp2 = max(d[i + 1][j], d[i + 1][j + 1]);
      dmax  = max(temp1, temp2);
      
      if(dmax < z[0] || dmin > z[nc-1])
	continue;
      
      for(k = 0; k < nc; k++) {
	if(z[k] < dmin || z[k] > dmax)
	  continue;
          
	for(m = 4; m >= 0; m--) {
	  if(m > 0) {
	    h[m]  = d[i + im[m - 1]][j + jm[m - 1]] - z[k];
	    xh[m] = x[i + im[m - 1]];
	    yh[m] = y[j + jm[m - 1]];
	  } else {
	    h[0]  = 0.25 * (h[1] + h[2] + h[3] + h[4]);
	    xh[0] = 0.50 * (x[i] +x[i + 1]);
	    yh[0] = 0.50 * (y[j] +y[j + 1]);
	  }
	  
	  if(h[m] > 0.0)
	    sh[m] = 1;
	  else if(h[m] < 0.0)
	    sh[m] = -1;
	  else
	    sh[m] = 0;
	}

	/*
	  Note: at this stage the relative heights of the corners and the
	  centre are in the h array, and the corresponding coordinates are
	  in the xh and yh arrays. The centre of the box is indexed by 0
	  and the 4 corners by 1 to 4 as shown below.
	  Each triangle is then indexed by the parameter m, and the 3
	  vertices of each triangle are indexed by parameters m1,m2,and m3.
	  It is assumed that the centre of the box is always vertex 2
	  though this isimportant only when all 3 vertices lie exactly on
	  the same contour level, in which case only the side of the box
	  is drawn.

	  vertex 4 +-------------------+ vertex 3
	           | \               / |
		   |   \    m-3    /   |
		   |     \       /     |
		   |       \   /       |
		   |  m=2    X   m=2   |       the centre is vertex 0
		   |       /   \       |
		   |     /       \     |
		   |   /    m=1    \   |
		   | /               \ |
	  vertex 1 +-------------------+ vertex 2
	
	*/
         
	/* Scan each triangle in the box */
	
	for(m = 1; m <= 4; m++) {
	  m1 = m;
	  m2 = 0;
	  
	  if(m != 4)
	    m3 = m + 1;
	  else
	    m3 = 1;
	  
	  if((case_value = castab[sh[m1] + 1][sh[m2] + 1][sh[m3] + 1]) == 0)
	    continue;
	  
	  switch(case_value) {
	  case 1: /* Line between vertices 1 and 2 */
	    x1 = xh[m1];
	    y1 = yh[m1];
	    x2 = xh[m2];
	    y2 = yh[m2];
	    break;
	  case 2: /* Line between vertices 2 and 3 */
	    x1 = xh[m2];
	    y1 = yh[m2];
	    x2 = xh[m3];
	    y2 = yh[m3];
	    break;
	  case 3: /* Line between vertices 3 and 1 */
	    x1 = xh[m3];
	    y1 = yh[m3];
	    x2 = xh[m1];
	    y2 = yh[m1];

	    break;
	  case 4: /* Line between vertex 1 and side 2-3 */
	    x1 = xh[m1];
	    y1 = yh[m1];
	    x2 = xsect(m2, m3);
	    y2 = ysect(m2, m3);

	    break;
	  case 5: /* Line between vertex 2 and side 3-1 */
	    x1 = xh[m2];
	    y1 = yh[m2];
	    x2 = xsect(m3, m1);
	    y2 = ysect(m3, m1);

	    break;
	  case 6: /* Line between vertex 3 and side 1-2 */
	    x1 = xh[m1];
	    y1 = yh[m1];
	    x2 = xsect(m1, m2);
	    y2 = ysect(m1, m2);
	    break;
	  case 7: /* Line between sides 1-2 and 2-3 */
	    x1 = xsect(m1, m2);
	    y1 = ysect(m1, m2);
	    x2 = xsect(m2, m3);
	    y2 = ysect(m2, m3);

	    break;
	  case 8: /* Line between sides 2-3 and 3-1 */
	    x1 = xsect(m2, m3);
	    y1 = ysect(m2, m3);
	    x2 = xsect(m3, m1);
	    y2 = ysect(m3, m1);

	    break;
	  case 9: /* Line between sides 3-1 and 1-2 */
	    x1 = xsect(m3, m1);
	    y1 = ysect(m3, m1);
	    x2 = xsect(m1, m2);
	    y2 = ysect(m1, m2);

	    break;
	  default:
	    break;
	  }
	  
	  /* Finally draw the line */
	
	  ConrecLine(x1, y1, x2, y2, z[k]);
	} /* m */
      } /* k - contour */
    } /* i */
  } /* j */
}

/* Table of constant values */

static double c_b3 = 2.0;
static double c_b4 = 4.0;

/* Main program */ 

int stream(double pm)
{
  /* Format strings */
  /* System generated locals */
  
  int i_1, i_2;
  double d_1, d_2, d_3;
  
  /* Builtin functions */
  
  extern double pow_dd();
  
  /* Local variables */
  
  static int i, j;
  static double dx, dy, nx, ny;
  
  /* Fortran I/O blocks */
  
  dx = 0.01;
  dy = 0.01;
  nx = (1.0 / dx);
  ny = (1.0 / dy);

  i_1 = (int) nx;
  
  for(i = 1; i <= i_1; ++i) {
    i_2 = (int) ny;

    for(j = 1; j <= i_2; ++j) {
      d_1 = (i * dx - 0.5);
      d_3 = (j * dy - 0.5);
      d_2 = (1.0 - pow_dd(&d_3, &c_b3) * 4.0);
      
      st(i, j) = - ((1.0 - pm) * 4.0 * pow_dd(&d_1, &c_b3) + pm) 
	* pow_dd(&d_2, &c_b4) * cos(M_PI * (i * dx - 0.5)); 
      
      //	* (1.0 / (4.0 * M_PI));
    }
  }
  
  return(0);
} /* MAIN */

double pow_dd(ap, bp) 
     double *ap, *bp;
{
  return(pow(*ap, *bp));
}






